#include "StdAfx.h"
#include ".\glslprogramobject.h"
#include <Log.h>

#define GLCHK CheckGLError(__FILE__, __LINE__)
//#define GLCHK 

static void CheckGLError(const char *fname, int line)
{
	int err = glGetError();
	if (err != GL_NO_ERROR)
	{
		Log.AddLine(__FILE__, "%s, %d, %s", fname, line, gluErrorString(err));
	}
}


static const char* LoadFile(const char* fname)
{
	FILE *f;
	f=fopen(fname, "rt");

	if (f==NULL) return NULL;

	int len; 
	fseek(f, 0, SEEK_END); len = ftell(f); fseek(f, 0, SEEK_SET);

	char* data = new char[len+1];
	memset(data, 0, len+1);
	fread(data, 1, len, f);

	fclose (f);
	return data;
}

static void printInfoLog(GLhandleARB object)
{
	int maxLength = 0;
	glGetObjectParameterivARB(object, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);

	char *infoLog = new char[maxLength];
	glGetInfoLogARB(object, maxLength, &maxLength, infoLog);

	Log.AddLine(__FILE__, "%s", infoLog);

	delete [] infoLog;
}


CGLSLProgramObject::CGLSLProgramObject(const char *name)
{
	m_handle = glCreateProgramObjectARB();
	m_bReady = false;
	m_Name = name;
}

CGLSLProgramObject::~CGLSLProgramObject(void)
{
	if (m_handle) glDeleteObjectARB(m_handle);
}

int CGLSLProgramObject::AttachShader(GLSLShaderObject* shader)
{
	m_ShaderFileNames.push_back(shader->m_Name);
	// attach vertex shader to program object
	glAttachObjectARB(m_handle, shader->object);

	// delete vertex object, no longer needed
	shader->Delete();

	GLenum err = glGetError();
	if (err != GL_NO_ERROR)
		Log.AddLine(__FILE__, "%s: OpenGL error: %s", m_Name.c_str(), gluErrorString(err));

	return 1;
}


int CGLSLProgramObject::Link()
{
	Log.AddLine(__FILE__, "Linking:");
	int i;
	for (i=0; i<m_ShaderFileNames.size(); i++)
	{
		Log.AddLine(__FILE__, "%... %s", m_ShaderFileNames[i].c_str());
	}

	glLinkProgramARB(m_handle);

	GLint linked = false;
	glGetObjectParameterivARB(m_handle, GL_OBJECT_LINK_STATUS_ARB, &linked);
	if (!linked)
	{
		Log.AddLine(__FILE__, "Shaders failed to link: %s", m_Name.c_str());
		printInfoLog(m_handle);
		return 0;
	}

	glValidateProgramARB(m_handle);

	GLint validated = false;
	glGetObjectParameterivARB(m_handle, GL_OBJECT_VALIDATE_STATUS_ARB, &validated);
	if (!validated)
	{
		Log.AddLine(__FILE__, "Shaders failed to validate: %s", m_Name.c_str());
		printInfoLog(m_handle);
		return 0;
	}

	GetUniforms();
	m_bReady = true;
	return 1;
}

int CGLSLProgramObject::Bind()
{
	if (m_bReady)
	{
		glUseProgramObjectARB(m_handle);
		//Log.AddLine(__FILE__, "=== Binding GLSL %s", m_Name.c_str());
		return 1;
	}
	else
	{
		Log.AddLine(__FILE__, "Can't bind program object %s", m_Name.c_str());
	}

	return 0;
}

int CGLSLProgramObject::Unbind()
{
	//Log.AddLine(__FILE__, "=== Unbind GLSL %s", m_Name.c_str());
	glUseProgramObjectARB(0);
	return 1;
}

typedef struct tagTypes
{
	int type;
	char *str;
} tTypes;

#define GLTYPE(_x) {_x, #_x}
tTypes glTypes[]=
{
	GLTYPE(GL_PROGRAM_OBJECT_ARB),
	GLTYPE(GL_SHADER_OBJECT_ARB),             
	GLTYPE(GL_OBJECT_TYPE_ARB),                
	GLTYPE(GL_OBJECT_SUBTYPE_ARB),             
	GLTYPE(GL_FLOAT_VEC2_ARB),                 
	GLTYPE(GL_FLOAT_VEC3_ARB),                 
	GLTYPE(GL_FLOAT_VEC4_ARB),                 
	GLTYPE(GL_INT_VEC2_ARB),                   
	GLTYPE(GL_INT_VEC3_ARB),                   
	GLTYPE(GL_INT_VEC4_ARB),                   
	GLTYPE(GL_BOOL_ARB),                       
	GLTYPE(GL_BOOL_VEC2_ARB),                  
	GLTYPE(GL_BOOL_VEC3_ARB),                  
	GLTYPE(GL_BOOL_VEC4_ARB),                  
	GLTYPE(GL_FLOAT_MAT2_ARB),                 
	GLTYPE(GL_FLOAT_MAT3_ARB),                 
	GLTYPE(GL_FLOAT_MAT4_ARB),                 

	{0, NULL}
};

char nista[] = "nistanistanistanistanistanistanista";

static const char *getGLType(int t)
{
	int i;
	for (i=0; glTypes[i].str!=0; i++)
	{
		if (t == glTypes[i].type)
			return glTypes[i].str;
	}

	sprintf(nista, "%04x", t);
	return nista;
};
		

void CGLSLProgramObject::Dump()
{
	int i;

	Log.AddLine(__FILE__, "Uniforms of %s", m_Name.c_str());
	for (i=0; i<m_Uniforms.size(); i++)
	{
		Log.AddLine(__FILE__, "%2d. %-32s %-20s %d", i, m_Uniforms[i].name.c_str(), getGLType(m_Uniforms[i].type), m_Uniforms[i].size);
	}
}

//GLAPI void APIENTRY glGetActiveUniformARB (GLhandleARB, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLcharARB *);

void CGLSLProgramObject::GetUniforms()
{
	GLcharARB uname[256];

	//
	// Example usage of GetActiveUniformARB.
	//
	GLint maxLength;
	GLint i, uniformCount;
	//
	// Get the number of uniforms, and the length of the longest name.
	//
	glGetObjectParameterivARB(m_handle, GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB, &maxLength);
	glGetObjectParameterivARB(m_handle, GL_OBJECT_ACTIVE_UNIFORMS_ARB, &uniformCount);
										
	//
	// Allocate arrays to store the answers in. For simplicity, the return from malloc is not checked
	// for NULL.
	//
	/*
	size = (GLint *) malloc(uniformCount * sizeof(GLint));
	type = (GLint *) malloc(uniformCount * sizeof(GLenum));
	length = (GLsizei *) malloc(uniformCount * sizeof(GLsizei));
	name = (GLcharARB **) malloc(uniformCount * sizeof(GLcharARB **));
	*/
	//
	// Loop over GetActiveUniformARB and store the results away.
	//
	for (i = 0; i < uniformCount; i++) 
	{
		//name[i] = (GLcharARB *) malloc(maxLength * sizeof(GLcharARB));
		GLsizei length;
		tGLSLUniform u;

		glGetActiveUniformARB(m_handle, i, maxLength, &length, &u.size, &u.type, uname);
		u.name = uname;
		m_Uniforms.push_back(u);
		//Log.AddLine(__FILE__, "Uniform %s", u.name.c_str());
	}
}

void CGLSLProgramObject::sendUniform(const char *name, float x)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform1fARB(loc, x);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, float x, float y)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform2fARB(loc, x, y);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, float x, float y, float z)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform3fARB(loc, x, y, z);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, float x, float y, float z, float w)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform4fARB(loc, x, y, z, w);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, int x)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform1iARB(loc, x);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, int x, int y)
{
	glUniform2iARB(getLoc(name), x, y);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, int x, int y, int z)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform3iARB(loc, x, y, z);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, int x, int y, int z, int w)
{
	int loc = getLoc(name);
	if (loc !=-1) glUniform4iARB(loc, x, y, z, w);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(const char *name, float *matrix, bool transpose, int size, int count)
{
	int loc = getLoc(name);
	if (loc != -1) 
	{
		switch(size)
		{
		case 2:
			glUniformMatrix2fvARB(loc, count, transpose, matrix);
			break;
		case 3:
			glUniformMatrix3fvARB(loc, count, transpose, matrix);
			break;
		case 4:
			glUniformMatrix4fvARB(loc, count, transpose, matrix);
			break;
		}
	} // if (loc != -1)
	GLCHK;
}




void CGLSLProgramObject::sendUniform(int loc, float x)
{
	if (loc !=-1) glUniform1fARB(loc, x);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, float x, float y)
{
	if (loc !=-1) glUniform2fARB(loc, x, y);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, float x, float y, float z)
{
	if (loc !=-1) glUniform3fARB(loc, x, y, z);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, float x, float y, float z, float w)
{
	if (loc !=-1) glUniform4fARB(loc, x, y, z, w);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, int x)
{
	if (loc !=-1) glUniform1iARB(loc, x);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, int x, int y)
{
	if (loc !=-1) glUniform2iARB(loc, x, y);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, int x, int y, int z)
{
	if (loc !=-1) glUniform3iARB(loc, x, y, z);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, int x, int y, int z, int w)
{
	if (loc !=-1) glUniform4iARB(loc, x, y, z, w);
	GLCHK;
}

void CGLSLProgramObject::sendUniform(int loc, float *matrix, bool transpose, int size, int count)
{
	if (loc != -1) 
	{
		switch(size)
		{
		case 2:
			glUniformMatrix2fvARB(loc, count, transpose, matrix);
			break;
		case 3:
			glUniformMatrix3fvARB(loc, count, transpose, matrix);
			break;
		case 4:
			glUniformMatrix4fvARB(loc, count, transpose, matrix);
			break;
		}
	} // if (loc != -1)
	GLCHK;
}


// Private part :

int CGLSLProgramObject::getLoc(const char *name)
{
	int loc = glGetUniformLocationARB(m_handle, name);
	if(loc == -1)
	{
		Log.AddLine(__FILE__, "No such uniform %s", name);
		return -1;
	}
	return loc;
}














GLSLShaderObject::GLSLShaderObject(void)
{
}

GLSLShaderObject::~GLSLShaderObject(void)
{

}

int GLSLShaderObject::LoadFromFile(const char* fname, GLint type)
{
	const char types_32bit[]=
		"typedef float Tfloat; \n"
		"typedef vec2 Tvec2;  \n"
		"typedef vec3 Tvec3;  \n"
		"typedef vec4 Tvec4;  \n"
		"typedef mat3 Tmat3;  \n"
		"typedef mat4 Tmat4;  \n\n";

	/*
	const char types_16bit[]=
		"typedef half  Tfloat; \n"
		"typedef half2 Tvec2;  \n"
		"typedef half3 Tvec3;  \n"
		"typedef half4 Tvec4;  \n"
		"typedef mat3  Tmat3;  \n"
		"typedef mat4  Tmat4;  \n\n";
	*/
	const char types_16bit[]=
		"#ifndef __GLSL_CG_DATA_TYPES\n"
		"# define half float\n"
		"# define half2 vec2\n"
		"# define half3 vec3\n"
		"# define half4 vec4\n"
		"#endif\n\n";

	m_Name = fname;
	object = glCreateShaderObjectARB(type);
	//assert(object != 0);

	const char* src = LoadFile(fname);
	if (src==NULL) 
	{
		Log.AddLine(__FILE__, "Error loading file %s", fname);
		return 0;
	}

	GLint length = (GLint)strlen(src);

	const GLcharARB*  progs[]    = { (GLcharARB*)types_16bit, (GLcharARB*)src};
	const GLint       progLens[] = {(GLint)strlen(types_16bit), length};

	glShaderSourceARB(object, 1, &src, &length);
	//glShaderSourceARB(object, 2, progs, progLens);
	glCompileShaderARB(object);

	delete [] src;

	// check if shader compiled
	GLint compiled = 0;
	glGetObjectParameterivARB(object, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);

	if (!compiled)
	{
		Log.AddLine(__FILE__, "glCompileShaderARB FAIL: %s", fname);
		printInfoLog(object);
		return 0;
	}

	return 1;
}

void GLSLShaderObject::Delete()
{
	glDeleteObjectARB(object);
}

